"""
    Copyright (C) 2024  Guo Zhengyang

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import cupy as cp
from Bio.PDB import *
import cv2

def find_neighbor(vector):
    nlst = []
    for i in range(len(vector)):
        v = vector.copy()
        if v[i] != 0:
          v[i] -= 1
          nlst.append(int(''.join([str(n) for n in v]),5))
        v = vector.copy()
        if v[i] != 4:
          v[i] += 1
          nlst.append(int(''.join([str(n) for n in v]),5))
    return nlst

def find_10base_neighbor(number):
  nlst = []
  for i in range(127):
      nlst.append(number+5**i)
      nlst.append(number-5**i)
  nlst = [n for n in nlst if n >= 0]
  return nlst


def categorize_matrix(mat,mode='embed'):
    cat = np.full(mat.shape,4)
    cat[(mat>np.cos(0.2*np.pi))&(mat<=np.cos(0.1*np.pi))] = 3
    cat[(mat>np.cos(0.3*np.pi))&(mat<=np.cos(0.2*np.pi))] = 2
    cat[(mat>np.cos(0.4*np.pi))&(mat<=np.cos(0.3*np.pi))] = 1
    cat[(mat>0)&(mat<=np.cos(0.4*np.pi))] = 0
    cat = cat[:,:-1]
    label = []
    for r in cat:
        fb = [int(n) for n in list(r)]
        if mode == 'embed':
            l = int(''.join([str(n) for n in fb]),5)
            label.append(l)
        elif mode == 'search':
            l = int(''.join([str(n) for n in fb]),5)
            label.extend(find_10base_neighbor(l))
            label.append(l)
    return label

def get_coordinate(pdb_path, name):
    p = PDBParser()
    st = p.get_structure(name, pdb_path)
    ca = np.array([atom.coord for atom in st.get_atoms() if atom.name == 'CA'])
    return ca


def extract_features(mat):
    mat = cv2.normalize(mat, None, 0, 255, cv2.NORM_MINMAX).astype('uint8')
    sift = cv2.SIFT_create()
    kp, des = sift.detectAndCompute(mat, None)
    return des


def norm(x):
    x = x.T
    x1 = np.linalg.norm(x=x, ord=2, axis=0, keepdims=True)
    x = x / x1
    return x.T

def compare(des, t, data, length):
    try:
        aa = int(length[data[0]])
        dist = cp.dot(des, data[1].T)
        max_mat = (cp.max(dist, axis=1) - t)
        ind = max_mat > 0
        n = cp.sum(max_mat * (max_mat > 0))
        match = int(cp.sum(ind))
        n = float(n)
        feat_q = des.shape[0]
        feat_k = data[1].shape[0]
        if 'CELE_' in data[0]:
            name = data[0].split('_')[1]
        else:
            name = data[0]
        return [name,match,n,feat_q,feat_k,aa]
    except:
        return ['error',0,0,0,0,0]

def z_score(arr):
    arr = np.asarray(arr, dtype=float)
    return (arr-np.mean(arr))/np.std(arr)


"""
    Copyright (C) 2024  Guo Zhengyang

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import pandas as pd
from adams.tool_kit import *
import cupy as cp
from scipy.spatial import distance_matrix as dm
import numpy as np
import os
import pickle as pkl
import GPUtil
import subprocess
from tqdm import tqdm

def get_gpu_info():
    gpus = GPUtil.getGPUs()
    return gpus

class ADAMS_match():
    def __init__(self, ref, gpu_usage, threshold=0.95):
        self.ref = ref
        self.features = cp.array([])
        self.distance_matrix = np.array([])
        self.label = []
        self.threshold = threshold
        self.extract_features()
        self.gpu_usage = gpu_usage
        self.total_task = 0
        self.chunk_size = 0
        self.process = []
        self.raw_result = []

    def extract_features(self):
        ca = get_coordinate(self.ref, 'X')
        dist = dm(ca, ca)
        features = extract_features(dist)
        features = norm(features)
        self.distance_matrix = dist
        self.features = cp.array(features)
        self.label = set(categorize_matrix(np.array(features),mode='embed')) ##
    def match(self, db_folder, temp, prefilter_threshold=0.01):
        if not os.path.exists(temp):
            os.mkdir(temp)
        script_path = subprocess.run('which compare_all.py',text=True,shell=True,capture_output=True).stdout[:-1]
        with open('./feature.pkl','wb') as f:
            pkl.dump([self.features,self.label],f)
        gpus = get_gpu_info()
        gpu_tasks = [int(n.memoryFree / 2000)  for n in gpus]
        for i in self.gpu_usage:
            self.total_task += gpu_tasks[i]
        files = os.listdir(db_folder)
        self.chunk_size = int(len(files) / self.total_task) + 1
        total = 0
        for i in self.gpu_usage:
            gpu = i
            max_task = gpu_tasks[i]
            current_task = 0
            for j in range(gpu_tasks[i]):
                start = total * self.chunk_size
                if current_task < max_task:
                    process = subprocess.Popen(
                        f'python {script_path} --ver {total} --feature_file ./feature.pkl --db_folder {db_folder} --ind1 {start} --ind2 {start + self.chunk_size} --device {gpu} --out_folder {temp} --prefilter {prefilter_threshold}',shell=True,text=True)
                    self.process.append(process)
                total += 1
        for i in self.process:
            i.wait()
        name = []
        match = []
        score = []
        identity = []
        files = os.listdir(temp)
        for i in files:
            if i[-4:] == '.pkl':
                with open(os.path.join(temp, i), 'rb') as f:
                    data = pkl.load(f)
                name.extend([n[0] for n in data])
                match.extend([n[1] for n in data])
                score.extend([n[2] for n in data])
                id_lst = []
                for n in data:
                    if n[3] != n[1]:
                      id_lst.append(n[1]/(n[3]-n[1]))
                    else:
                      id_lst.append(1)
                identity.extend(id_lst)
        score_z = z_score(score)
        match_z = z_score(match)
        total_z = score_z + match_z

        result = pd.DataFrame(
                {'name': name, 'match': match, 'score': score, 'z_match': match_z, 'z_score': score_z, 'z': total_z,'identity':identity})
        os.system(f'rm -r {temp}'+'/*')
        result = result[result['match']!=0]
        result_sort = result.sort_values(by='z',ascending=False)
        return result_sort



"""
    Copyright (C) 2024  Guo Zhengyang

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from adams.tool_kit import *
import os
from scipy.spatial import distance_matrix
import cupy as cp
import pickle as pkl
from tqdm.contrib.concurrent import process_map
from functools import partial
import gzip


def get_feature_list(pdb_folder, out_folder, file):
    error = []
    try:
        ca = get_coordinate(os.path.join(pdb_folder, file), 'X')
        name = file[:-4]
        dist = distance_matrix(ca, ca)
        dist = extract_features(dist)
        dist = norm(dist)
        length = len(ca)
        label = set(categorize_matrix(np.vstack(dist),mode='embed'))
        # return name, dist, length
        with open(os.path.join(out_folder, file[:-3]+'pkl'), 'wb') as f:
            pkl.dump([name, dist, length, label], f)
    except:
        error.append('1')



class DatabaseMaker():
    def __init__(self, device=0, process=40):
        self.device = device
        self.process = process

    def make(self, pdb_folder, out_folder):
        mempool = cp.get_default_memory_pool()
        files = os.listdir(pdb_folder)
        feature_p = partial(get_feature_list, pdb_folder, out_folder)
        process_map(feature_p,files,max_workers=self.process,chunksize = int(len(files)/self.process))
        mempool.free_all_blocks()




"""
    Copyright (C) 2024  Guo Zhengyang

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import argparse
import pickle as pkl
import os
import cupy as cp
from tqdm import tqdm

parser = argparse.ArgumentParser()
parser.add_argument('--feature_file')
parser.add_argument('--ver')
parser.add_argument('--db_folder')
parser.add_argument('--ind1')
parser.add_argument('--ind2')
parser.add_argument('--device')
parser.add_argument('--prefilter')
parser.add_argument('--out_folder')
args = parser.parse_args()

db_folder, part, ind1, ind2, device, out_folder,feature_path,prefilter = args.db_folder, args.ver, int(args.ind1), int(args.ind2), int(
    args.device), args.out_folder, args.feature_file,float(args.prefilter)

def match(feature, label, path, prefilter):
    try:
        with open(path, 'rb') as f:
            data = pkl.load(f)
        label1 = data[-1]
        if len(label1&label) > prefilter*len(label):
          name = data[0]
          dt = data[1]
          len2 = dt.shape[0]
          dt = cp.array(dt)
          len1 = feature.shape[0]
          feature = cp.array(feature)
          res = cp.max(cp.dot(dt, feature.T) - 0.95, axis=0)
          ind = res > 0
          score = float(cp.sum(res * ind))
          match = int(cp.sum(ind))
          sumup = len1+len2
          return name, match, score, sumup
        else:
          return 0, 0, 0, 0
    except:
        return 0, 0, 0, 0


path = [os.path.join(db_folder,n) for n in os.listdir(db_folder)]
with open(feature_path, 'rb') as f:
    feature = pkl.load(f)
with cp.cuda.Device(device):
    label = feature[-1]
    feature = cp.array(feature[0])
    mempool = cp.get_default_memory_pool()
    res = []
    for j in path[ind1:ind2]:
        res.append(match(feature,label,j,prefilter))
    mempool.free_all_blocks()
    with open(out_folder + '/{}.pkl'.format(j.split('/')[-1].split('.')[0]), 'wb') as f:
        pkl.dump(res, f)
